# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Common functions for parsing input/output files or XMLschemas from FLEUR
"""
from lxml import etree


def clear_xml(tree):
    """
    Removes comments and executes xinclude tags of an
    xml tree.

    :param tree: an xml-tree which will be processed

    :returns: cleared_tree, an xmltree without comments and with replaced xinclude tags
    """
    import copy

    cleared_tree = copy.deepcopy(tree)

    #find any include tags
    include_tags = eval_xpath(cleared_tree,
                              '//xi:include',
                              namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                              list_return=True)

    parents = []
    known_tags = []
    for tag in include_tags:
        parent = tag.getparent()
        parents.append(parent)
        known_tags.append({elem.tag for elem in parent})

    # replace XInclude parts to validate against schema
    if len(include_tags) != 0:
        cleared_tree.xinclude()

    # get rid of xml:base attribute in the included parts
    for parent, old_tags in zip(parents, known_tags):
        new_tags = {elem.tag for elem in parent}

        #determine the elements not in old_tags, which are in tags
        #so what should have been included
        included_tag_names = new_tags.difference(old_tags)

        #Check for emtpy set (relax.xml include may not insert something)
        if not included_tag_names:
            continue

        for tag_name in included_tag_names:
            for elem in parent:
                if elem.tag == tag_name:
                    for attribute in elem.keys():
                        if 'base' in attribute:
                            try:
                                del elem.attrib[attribute]
                            except BaseException:
                                pass

    # remove comments from inp.xml
    comments = cleared_tree.xpath('//comment()')
    for comment in comments:
        com_parent = comment.getparent()
        com_parent.remove(comment)

    return cleared_tree


def validate_xml(xmltree, schema, error_header='File does not validate'):
    """
    Checks a given xmltree against a schema and produces a nice error message
    with all the validation errors collected

    :param xmltree: xmltree of the file to validate
    :param schema: etree.XMLSchema to validate against
    :param error_header: str to lead a evtl error message with

    :raises: etree.DocumentInvalid if the schema does not validate
    """
    from itertools import groupby

    try:
        schema.assertValid(clear_xml(xmltree))
    except etree.DocumentInvalid as exc:
        error_log = sorted(schema.error_log, key=lambda x: x.message)
        error_output = []
        first_occurence = []
        for message, group in groupby(error_log, key=lambda x: x.message):
            err_occurences = list(group)
            error_message = f'Line {err_occurences[0].line}: {message}'
            error_lines = ''
            if len(err_occurences) > 1:
                error_lines = f"; This error also occured on the lines {', '.join([str(x.line) for x in err_occurences[1:]])}"
            error_output.append(f'{error_message}{error_lines} \n')
            first_occurence.append(err_occurences[0].line)

        error_output = [line for _, line in sorted(zip(first_occurence, error_output))]
        errmsg = f"{error_header}: \n{''.join(error_output)}"
        raise etree.DocumentInvalid(errmsg) from exc


def convert_xml_attribute(stringattribute, possible_types, constants=None, conversion_warnings=None):
    """
    Tries to converts a given string attribute to the types given in possible_types.
    First succeeded conversion will be returned

    :param stringattribute: str, Attribute to convert.
    :param possible_types: list of str What types it will try to convert to
    :param constants: dict, of constants defined in fleur input
    :param conversion_warnings: dict with warings about failed conversions

    :return: The converted value of the first succesful conversion
    """
    from masci_tools.util.fleur_calculate_expression import calculate_expression

    if not isinstance(stringattribute, list):
        stringattribute = [stringattribute]

    if conversion_warnings is None:
        conversion_warnings = []

    converted_list = []
    all_success = True
    for attrib in stringattribute:
        suc = False
        for value_type in possible_types:
            if value_type == 'float':
                converted_value, suc = convert_to_float(attrib, conversion_warnings=conversion_warnings)
            elif value_type == 'float_expression':
                if constants is None:
                    raise ValueError(
                        "For calculating attributes of the type 'float_expression' constants have to be given")
                try:
                    converted_value = calculate_expression(attrib, constants)
                    suc = True
                except ValueError as errmsg:
                    suc = False
                    conversion_warnings.append(f"Could not evaluate expression '{attrib}' "
                                               f'The following error was raised: {errmsg}')
            elif value_type == 'int':
                converted_value, suc = convert_to_int(attrib, conversion_warnings=conversion_warnings)
            elif value_type == 'switch':
                converted_value, suc = convert_from_fortran_bool(attrib, conversion_warnings=conversion_warnings)
            elif value_type == 'string':
                suc = True
                converted_value = str(attrib)
            if suc:
                converted_list.append(converted_value)
                break
        if not suc:
            converted_list.append(attrib)
            all_success = False

    ret_value = converted_list
    if len(converted_list) == 1:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_attribute_to_xml(attributevalue, possible_types, conversion_warnings=None, float_format='.10'):
    """
    Tries to converts a given attributevalue to a string for a xml file according
    to the types given in possible_types.
    First succeeded conversion will be returned

    :param attributevalue: value to convert.
    :param possible_types: list of str What types it will try to convert from
    :param conversion_warnings: dict with warings about failed conversions

    :return: The converted str of the value of the first succesful conversion
    """

    if not isinstance(attributevalue, list):
        attributevalue = [attributevalue]

    if conversion_warnings is None:
        conversion_warnings = []

    possible_types = possible_types.copy()

    if 'int' in possible_types:  #Since it just converts to string
        possible_types.remove('int')
        possible_types.append('int')

    if 'string' in possible_types:
        possible_types.remove('string')
    possible_types.append('string')  #Always try string

    converted_list = []
    all_success = True
    for value in attributevalue:

        suc = False
        for value_type in possible_types:
            if value_type == 'float':
                try:
                    converted_value = f'{value:{float_format}f}'
                    suc = True
                except ValueError as errmsg:
                    suc = False
                    conversion_warnings.append(f"Could not convert to float string '{value}' "
                                               f'The following error was raised: {errmsg}')
            elif value_type == 'float_expression':
                try:
                    converted_value = f'{value:{float_format}f}'
                    suc = True
                except ValueError as errmsg:
                    suc = False
                    conversion_warnings.append(f"Could not convert to float string '{value}' "
                                               f'The following error was raised: {errmsg}')
            elif value_type == 'switch':
                converted_value, suc = convert_to_fortran_bool(value, conversion_warnings=conversion_warnings)
            elif value_type in ('string', 'int'):
                suc = True
                converted_value = str(value)
            if suc:
                converted_list.append(converted_value)
                break
        if not suc:
            converted_list.append(value)
            all_success = False

    ret_value = converted_list
    if len(converted_list) == 1:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_xml_text(tagtext, possible_definitions, constants=None, conversion_warnings=None):
    """
    Tries to converts a given string text based on the definitions (length and type).
    First succeeded conversion will be returned

    :param tagtext: str, text to convert.
    :param possible_defintions: list of dicts What types it will try to convert to
    :param constants: dict, of constants defined in fleur input
    :param conversion_warnings: dict with warings about failed conversions

    :return: The converted value of the first succesful conversion
    """

    if conversion_warnings is None:
        conversion_warnings = []

    if not isinstance(tagtext, list):
        tagtext = [tagtext]

    converted_list = []
    all_success = True
    for text in tagtext:

        base_text = text.strip()
        split_text = text.split(' ')
        while '' in split_text:
            split_text.remove('')

        text_definition = None
        for definition in possible_definitions:
            if definition['length'] == len(split_text):
                text_definition = definition

        if text_definition is None:
            for definition in possible_definitions:
                if definition['length'] == 'unbounded' or \
                   definition['length'] == 1:
                    text_definition = definition

        if text_definition is None:
            conversion_warnings.append(f"Failed to convert '{text}', no matching definition found ")
            converted_list.append(text)
            all_success = False
            continue

        #Avoid splitting the text accidentally if there is no length restriction
        if text_definition['length'] == 1:
            split_text = [base_text]

        converted_text = []
        for value in split_text:
            warnings = []
            converted_value, suc = convert_xml_attribute(value,
                                                         text_definition['type'],
                                                         constants=constants,
                                                         conversion_warnings=warnings)
            converted_text.append(converted_value)
            if not suc:
                all_success = False
                for warning in warnings:
                    conversion_warnings.append(warning)

        if len(converted_text) == 1 and text_definition['length'] != 'unbounded':
            converted_text = converted_text[0]

        converted_list.append(converted_text)

    ret_value = converted_list
    if len(converted_list) == 1:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_text_to_xml(textvalue, possible_definitions, conversion_warnings=None, float_format='16.13'):
    """
    Tries to convert a given list of values to str for a xml file based on the definitions (length and type).
    First succeeded conversion will be returned

    :param textvalue: value to convert
    :param possible_definitions: list of dicts What types it will try to convert to
    :param conversion_warnings: dict with warings about failed conversions

    :return: The converted value of the first succesful conversion
    """

    if conversion_warnings is None:
        conversion_warnings = []

    if not isinstance(textvalue, list):
        textvalue = [textvalue]
    elif not isinstance(textvalue[0], list):
        textvalue = [textvalue]

    converted_list = []
    all_success = True
    for text in textvalue:

        if not isinstance(text, list):
            text = [text]

        text_definition = None
        for definition in possible_definitions:
            if definition['length'] == len(text):
                text_definition = definition

        if text_definition is None:
            for definition in possible_definitions:
                if definition['length'] == 'unbounded':
                    text_definition = definition

        if text_definition is None:
            conversion_warnings.append(f"Failed to convert '{text}', no matching definition found ")
            if isinstance(text, str):
                converted_list.append(text)
                continue
            converted_list.append('')
            all_success = False
            continue

        converted_text = []
        for value in text:
            warnings = []
            converted_value, suc = convert_attribute_to_xml(value,
                                                            text_definition['type'],
                                                            conversion_warnings=warnings,
                                                            float_format=float_format)
            converted_text.append(converted_value)
            if not suc:
                if isinstance(value, str):
                    converted_list.append(value)
                    continue
                all_success = False
                for warning in warnings:
                    conversion_warnings.append(warning)

        converted_list.append(' '.join(converted_text))

    ret_value = converted_list
    if len(converted_list) == 1:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_to_float(value_string, conversion_warnings=None):
    """
    Tries to make a float out of a string. If it can't it logs a warning
    and returns True or False if convertion worked or not.

    :param value_string: a string
    :param conversion_warnings: list with warings about failed conversions

    :return: value the new float or value_string: the string given
    :return: True if convertion was successfull, False otherwise
    """
    if conversion_warnings is None:
        conversion_warnings = []
    try:
        value = float(value_string)
    except TypeError:
        conversion_warnings.append(f"Could not convert: '{value_string}' to float, TypeError")
        return value_string, False

    except ValueError:
        conversion_warnings.append(f"Could not convert: '{value_string}' to float, ValueError")
        return value_string, False

    return value, True


def convert_to_int(value_string, conversion_warnings=None):
    """
    Tries to make a int out of a string. If it can't it logs a warning
    and returns True or False if convertion worked or not.

    :param value_string: a string
    :param conversion_warnings: list with warings about failed conversions

    :return: value the new int or value_string: the string given
    :return: True or False
    """
    if conversion_warnings is None:
        conversion_warnings = []
    try:
        value = int(value_string)
    except TypeError:
        conversion_warnings.append(f"Could not convert: '{value_string}' to int, TypeError")
        return value_string, False

    except ValueError:
        conversion_warnings.append(f"Could not convert: '{value_string}' to int, ValueError")
        return value_string, False

    return value, True


def convert_from_fortran_bool(stringbool, conversion_warnings=None):
    """
    Converts a string in this case ('T', 'F', or 't', 'f') to True or False

    :param stringbool: a string ('t', 'f', 'F', 'T')
    :param conversion_warnings: list with warings about failed conversions

    :return: boolean  (either True or False)
    """
    if conversion_warnings is None:
        conversion_warnings = []
    true_items = ['True', 't', 'T']
    false_items = ['False', 'f', 'F']
    if isinstance(stringbool, str):
        if stringbool in false_items:
            converted_value = False
            suc = True
        elif stringbool in true_items:
            converted_value = True
            suc = True
        else:
            suc = False
            converted_value = stringbool
            conversion_warnings.append(f"Could not convert: '{stringbool}' to boolean, "
                                       "which is not 'True', 'False', 't', 'T', 'F' or 'f'")
    elif isinstance(stringbool, bool):
        converted_value = stringbool  # no conversion needed...
        suc = True
    else:
        suc = False
        converted_value = stringbool
        conversion_warnings.append(f"Could not convert: '{stringbool}' to boolean, " 'only accepts str or boolean')

    return converted_value, suc


def convert_to_fortran_bool(boolean, conversion_warnings=None):
    """
    Converts a Boolean as string to the format defined in the input

    :param boolean: either a boolean or a string ('True', 'False', 'F', 'T')
    :param conversion_warnings: list with warings about failed conversions

    :return: a string (either 't' or 'f')
    """
    if conversion_warnings is None:
        conversion_warnings = []

    suc = False
    new_string = None
    if isinstance(boolean, bool):
        if boolean:
            new_string = 'T'
            suc = True
        else:
            new_string = 'F'
            suc = True
    elif isinstance(boolean, str):  # basestring):
        if boolean in ('True', 't', 'T'):
            new_string = 'T'
            suc = True
        elif boolean in ('False', 'f', 'F'):
            new_string = 'F'
            suc = True
        else:
            suc = False
            conversion_warnings.append(f"A string: {boolean} for a boolean was given, which is not 'True',"
                                       "'False', 't', 'T', 'F' or 'f'")
    else:
        suc = False
        conversion_warnings.append('convert_to_fortran_bool accepts only a string or '
                                   f'bool as argument, given {boolean} ')

    return new_string, suc


def eval_xpath(node, xpath, parser_info_out=None, list_return=False, namespaces=None):
    """
    Tries to evaluate an xpath expression. If it fails it logs it.
    If a absolute path is given (starting with '/') and the tag of the node
    does not match the root.
    It will try to find the tag in the path and convert it into a relative path

    :param node: root node of an etree
    :param xpath: xpath expression (relative, or absolute)
    :param parser_info_out: dict with warnings, info, errors, ...
    :param list_return: if True, the returned quantity is always a list even if only one element is in it
    :param namespaces: dict, passed to namespaces argument in xpath call

    :returns: text, attribute or a node list
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    assert isinstance(node, (etree._Element, etree._ElementTree)), f'Wrong Type for xpath eval; Got: {type(node)}'

    if isinstance(node, etree._Element):
        if node.tag != xpath.split('/')[1] and xpath.split('/')[0] != '.':
            #absolute path with a different root tag than node
            if node.tag in xpath:
                if '@' not in xpath:
                    xpath = xpath + '/'
                xpath = xpath.replace('/'.join(xpath.split(node.tag + '/')[:-1]) + node.tag, '.')
                xpath = xpath.rstrip('/')

    try:
        if namespaces is not None:
            return_value = node.xpath(xpath, namespaces=namespaces)
        else:
            return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        parser_info_out['parser_warnings'].append(f'There was a XpathEvalError on the xpath: {xpath} \n Either it does '
                                                  'not exist, or something is wrong with the expression.')
        return []  # or rather None?
    if len(return_value) == 1 and not list_return:
        return return_value[0]
    else:
        return return_value


def get_xml_attribute(node, attributename, parser_info_out=None):
    """
    Get an attribute value from a node.

    :param node: a node from etree
    :param attributename: a string with the attribute name.
    :param parser_info_out: dict with warnings, info, errors, ...
    :returns: either attributevalue, or None
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    if etree.iselement(node):
        attrib_value = node.get(attributename)
        if attrib_value:
            return attrib_value
        else:
            parser_info_out['parser_warnings'].append('Tried to get attribute: "{}" from element {}.\n '
                                                      'I recieved "{}", maybe the attribute does not exist'
                                                      ''.format(attributename, node.tag, attrib_value))
            return None
    else:  # something doesn't work here, some nodes get through here
        parser_info_out['parser_warnings'].append('Can not get attributename: "{}" from node of type {}, '
                                                  'because node is not an element of etree.'
                                                  ''.format(attributename, type(node)))
        return None


def split_off_tag(xpath):
    """
    Splits off the last part of the given xpath

    :param xpath: str of the xpath to split up
    """
    split_xpath = xpath.split('/')
    if split_xpath[-1] == '':
        return '/'.join(split_xpath[:-2]), split_xpath[-2]
    else:
        return '/'.join(split_xpath[:-1]), split_xpath[-1]


def split_off_attrib(xpath):
    """
    Splits off attribute of the given xpath (part after @)

    :param xpath: str of the xpath to split up
    """
    split_xpath = xpath.split('/@')
    assert len(split_xpath) == 2, f"Splitting off attribute failed for: '{split_xpath}'"
    return tuple(split_xpath)


def check_complex_xpath(node, base_xpath, complex_xpath):
    """
    Check that the given complex xpath produces a subset of the results
    for the simple xpath

    :param node: root node of an etree
    :param base_xpath: str of the xpath without complex syntax
    :param complex_xpath: str of the xpath to check

    :raises ValueError: If the complex_xpath does not produce a subset of the results
                        of the base_xpath
    """

    results_base = set(eval_xpath(node, base_xpath, list_return=True))
    results_complex = set(eval_xpath(node, complex_xpath, list_return=True))

    if not results_base.issuperset(results_complex):
        raise ValueError(f"Complex xpath '{complex_xpath}' is not compatible with the base_xpath '{base_xpath}'")



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains helper functions for extracting information easily from the
schema_dicts defined for the Fleur input/output

Also provides convienient functions to use just a attribute name for extracting the
attribute from the right place in the given etree
"""
from masci_tools.util.parse_tasks_decorators import register_parsing_function
from masci_tools.util.lockable_containers import LockableList
from lxml import etree


def get_tag_xpath(schema_dict, name, contains=None, not_contains=None):
    """
    Tries to find a unique path from the schema_dict based on the given name of the tag
    and additional further specifications

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path

    :returns: str, xpath to the given tag

    :raises ValueError: If no unique path could be found
    """

    if contains is None:
        contains = []
    elif not isinstance(contains, (list, set)):
        contains = [contains]

    if not_contains is None:
        not_contains = []
    elif not isinstance(not_contains, (list, set)):
        not_contains = [not_contains]

    possible_lists = ['tag_paths']

    if 'iteration_tag_paths' in schema_dict:
        possible_lists += ['iteration_tag_paths']

    all_paths = []
    for list_name in possible_lists:
        if name in schema_dict[list_name]:
            paths = schema_dict[list_name][name]

            if not isinstance(paths, LockableList):
                paths = [paths]
            else:
                paths = paths.get_unlocked()

            invalid_paths = set()
            for phrase in contains:
                for xpath in paths:
                    if phrase not in xpath:
                        invalid_paths.add(xpath)

            for phrase in not_contains:
                for xpath in paths:
                    if phrase in xpath:
                        invalid_paths.add(xpath)

            for invalid in invalid_paths:
                paths.remove(invalid)

            all_paths += paths

    if len(all_paths) == 1:
        return all_paths[0]
    elif len(all_paths) == 0:
        raise ValueError(f'The tag {name} has no possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}')
    else:
        raise ValueError(f'The tag {name} has multiple possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains} \n'
                         f'These are possible: {all_paths}')


def get_tag_info(schema_dict, name, contains=None, not_contains=None, path_return=True, convert_to_builtin=False):
    """
    Tries to find a unique path from the schema_dict based on the given name of the tag
    and additional further specifications and returns the tag_info entry for this tag

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path
    :param path_return: bool, if True the found path will be returned alongside the tag_info
    :param convert_to_builtin: bool, if True the CaseInsensitiveFrozenSets are converetd to normal sets
                               with the rigth case of the attributes

    :returns: dict, tag_info for the found xpath
    :returns: str, xpath to the tag if `path_return=True`
    """
    import copy
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    tag_xpath = get_tag_xpath(schema_dict, name, contains=contains, not_contains=not_contains)
    tag_info = copy.deepcopy(schema_dict['tag_info'][tag_xpath])

    if convert_to_builtin:
        tag_info = {
            key: set(val.original_case.values()) if isinstance(val, CaseInsensitiveFrozenSet) else val
            for key, val in tag_info.items()
        }

    if path_return:
        return tag_info, tag_xpath
    else:
        return tag_info


def get_attrib_xpath(schema_dict, name, contains=None, not_contains=None, exclude=None, tag_name=None):
    """
    Tries to find a unique path from the schema_dict based on the given name of the attribute
    and additional further specifications

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path
    :param exclude: list of str, here specific types of attributes can be excluded
                    valid values are: settable, settable_contains, other
    :param tag_name: str, if given this name will be used to find a path to a tag with the
                     same name in :py:func:`get_tag_xpath()`

    :returns: str, xpath to the tag with the given attribute

    :raises ValueError: If no unique path could be found
    """

    if tag_name is not None:
        tag_xpath = get_tag_xpath(schema_dict, tag_name, contains=contains, not_contains=not_contains)

        err_msg = f'No attribute {name} found at tag {tag_name}'
        if tag_xpath in schema_dict['tag_info']:
            if name not in schema_dict['tag_info'][tag_xpath]['attribs']:
                raise ValueError(err_msg)
        else:
            if name not in schema_dict['iteration_tag_info'][tag_xpath]['attribs']:
                raise ValueError(err_msg)
        return f'{tag_xpath}/@{name}'

    if contains is None:
        contains = []
    elif not isinstance(contains, (list, set)):
        contains = [contains]

    if not_contains is None:
        not_contains = []
    elif not isinstance(not_contains, (list, set)):
        not_contains = [not_contains]

    possible_lists = ['unique_attribs', 'unique_path_attribs', 'other_attribs']
    output = False
    if 'iteration_unique_attribs' in schema_dict:
        #outputschema
        output = True
        possible_lists += ['iteration_unique_attribs', 'iteration_unique_path_attribs', 'iteration_other_attribs']

    if exclude is not None:
        for list_name in exclude:
            possible_lists.remove(f'{list_name}_attribs')
            if output:
                possible_lists.remove(f'iteration_{list_name}_attribs')
    all_paths = []
    for list_name in possible_lists:
        if name in schema_dict[list_name]:
            paths = schema_dict[list_name][name]

            if not isinstance(paths, LockableList):
                paths = [paths]
            else:
                paths = paths.get_unlocked()

            invalid_paths = set()
            for phrase in contains:
                for xpath in paths:
                    if phrase not in xpath:
                        invalid_paths.add(xpath)

            for phrase in not_contains:
                for xpath in paths:
                    if phrase in xpath:
                        invalid_paths.add(xpath)

            for invalid in invalid_paths:
                paths.remove(invalid)

            all_paths += paths

    if len(all_paths) == 1:
        return all_paths[0]
    elif len(all_paths) == 0:
        raise ValueError(f'The attrib {name} has no possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}')
    else:
        raise ValueError(f'The attrib {name} has multiple possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}\n'
                         f'These are possible: {all_paths}')


def read_constants(root, schema_dict):
    """
    Reads in the constants defined in the inp.xml
    and returns them combined with the predefined constants from
    fleur as a dictionary

    :param root: root of the etree of the inp.xml file
    :param schema_dict: schema_dictionary of the version of the file to read (inp.xml or out.xml)

    :return: a python dictionary with all defined constants
    """
    from masci_tools.util.constants import FLEUR_DEFINED_CONSTANTS
    import copy

    defined_constants = copy.deepcopy(FLEUR_DEFINED_CONSTANTS)

    constants = evaluate_tag(root, schema_dict, 'constant', defined_constants)

    if constants['name'] is not None:
        if not isinstance(constants['name'], list):
            constants = {key: [val] for key, val in constants.items()}
        for name, value in zip(constants['name'], constants['value']):
            if name not in defined_constants:
                defined_constants[name] = value
            else:
                raise KeyError(f'Ambiguous definition of key {name}')

    return defined_constants


@register_parsing_function('attrib')
def evaluate_attribute(node, schema_dict, name, constants=None, parser_info_out=None, **kwargs):
    """
    Evaluates the value of the attribute based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param constants: dict, contains the defined constants
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: list or single value, converted in convert_xml_attribute
    """
    from masci_tools.util.xml.common_xml_util import eval_xpath, convert_xml_attribute

    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            kwargs['contains'] = set(kwargs.get('contains', []))
            kwargs['contains'].add(node.tag)

    attrib_xpath = get_attrib_xpath(schema_dict, name, **kwargs)

    stringattribute = eval_xpath(node, attrib_xpath, parser_info_out=parser_info_out)

    if isinstance(stringattribute, list):
        if len(stringattribute) == 0:
            parser_info_out['parser_warnings'].append(f'No values found for attribute {name}')
            return None

    possible_types = schema_dict['attrib_types'][name]

    warnings = []
    converted_value, suc = convert_xml_attribute(stringattribute,
                                                 possible_types,
                                                 constants=constants,
                                                 conversion_warnings=warnings)

    if not suc:
        parser_info_out['parser_warnings'].append(f'Failed to evaluate attribute {name}: '
                                                  'Below are the warnings from convert_xml_attribute')
        for warning in warnings:
            parser_info_out['parser_warnings'].append(warning)

    return converted_value


@register_parsing_function('text')
def evaluate_text(node, schema_dict, name, constants, parser_info_out=None, **kwargs):
    """
    Evaluates the text of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path


    :returns: list or single value, converted in convert_xml_text
    """
    from masci_tools.util.xml.common_xml_util import eval_xpath, convert_xml_text

    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            kwargs['contains'] = set(kwargs.get('contains', []))
            kwargs['contains'].add(node.tag)

    tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    stringtext = eval_xpath(node, f'{tag_xpath}/text()', parser_info_out=parser_info_out)

    if isinstance(stringtext, list):
        for text in stringtext.copy():
            if text.strip() == '':
                stringtext.remove(text)
    else:
        if stringtext.strip() == '':
            stringtext = []

    if isinstance(stringtext, list):
        if len(stringtext) == 0:
            parser_info_out['parser_warnings'].append(f'No text found for tag {name}')
            return None

    possible_definitions = schema_dict['simple_elements'][name]

    warnings = []
    converted_value, suc = convert_xml_text(stringtext,
                                            possible_definitions,
                                            constants=constants,
                                            conversion_warnings=warnings)

    if not suc:
        parser_info_out['parser_warnings'].append(f'Failed to evaluate text for tag {name}: '
                                                  'Below are the warnings from convert_xml_text')
        for warning in warnings:
            parser_info_out['parser_warnings'].append(warning)

    return converted_value


@register_parsing_function('allAttribs', all_attribs_keys=True)
def evaluate_tag(node, schema_dict, name, constants=None, parser_info_out=None, **kwargs):
    """
    Evaluates all attributes of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse

    :returns: dict, with attribute values converted via convert_xml_attribute
    """
    from masci_tools.util.xml.common_xml_util import eval_xpath, convert_xml_attribute

    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    only_required = kwargs.pop('only_required', False)
    ignore = kwargs.pop('ignore', None)

    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            kwargs['contains'] = set(kwargs.get('contains', []))
            kwargs['contains'].add(node.tag)

    tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    #Which attributes are expected
    attribs = set()
    if tag_xpath in schema_dict['tag_info']:
        attribs = schema_dict['tag_info'][tag_xpath]['attribs']
        optional = schema_dict['tag_info'][tag_xpath]['optional_attribs']
    elif 'iteration_tag_info' in schema_dict:
        if tag_xpath in schema_dict['iteration_tag_info']:
            attribs = schema_dict['iteration_tag_info'][tag_xpath]['attribs']
            optional = schema_dict['iteration_tag_info'][tag_xpath]['optional_attribs']

    if only_required:
        attribs = attribs.difference(optional)

    if ignore:
        attribs = attribs.difference(ignore)

    if not attribs:
        parser_info_out['parser_warnings'].append(f'Failed to evaluate attributes from tag {name}: '
                                                  'No attributes to parse either the tag does not '
                                                  'exist or it has no attributes')
    else:
        attribs = sorted(list(attribs.original_case.values()))

    out_dict = {}

    for attrib in attribs:

        stringattribute = eval_xpath(node, f'{tag_xpath}/@{attrib}', parser_info_out=parser_info_out)

        if isinstance(stringattribute, list):
            if len(stringattribute) == 0:
                parser_info_out['parser_warnings'].append(f'No values found for attribute {attrib} at tag {name}')
                out_dict[attrib] = None
                continue

        possible_types = schema_dict['attrib_types'][attrib]

        warnings = []
        out_dict[attrib], suc = convert_xml_attribute(stringattribute,
                                                      possible_types,
                                                      constants=constants,
                                                      conversion_warnings=warnings)

        if not suc:
            parser_info_out['parser_warnings'].append(f'Failed to evaluate attribute {attrib}: '
                                                      'Below are the warnings from convert_xml_attribute')
            for warning in warnings:
                parser_info_out['parser_warnings'].append(warning)

    return out_dict


@register_parsing_function('singleValue', all_attribs_keys=True)
def evaluate_single_value_tag(node, schema_dict, name, constants=None, parser_info_out=None, **kwargs):
    """
    Evaluates the value and unit attribute of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse

    :returns: value and unit, both converted in convert_xml_attribute
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    only_required = kwargs.get('only_required', False)

    value_dict = evaluate_tag(node, schema_dict, name, constants=constants, parser_info_out=parser_info_out, **kwargs)

    if 'value' not in value_dict:
        parser_info_out['parser_warnings'].append(f'Failed to evaluate singleValue from tag {name}: '
                                                  "Has no 'value' attribute")
    if 'units' not in value_dict and not only_required:
        parser_info_out['parser_warnings'].append(f'Failed to evaluate singleValue from tag {name}: '
                                                  "Has no 'units' attribute")

    return value_dict


@register_parsing_function('parentAttribs', all_attribs_keys=True)
def evaluate_parent_tag(node, schema_dict, name, constants=None, parser_info_out=None, **kwargs):
    """
    Evaluates all attributes of the parent tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse

    :returns: dict, with attribute values converted via convert_xml_attribute
    """
    from masci_tools.util.xml.common_xml_util import eval_xpath, convert_xml_attribute, get_xml_attribute

    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    only_required = kwargs.pop('only_required', False)
    ignore = kwargs.pop('ignore', None)

    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            kwargs['contains'] = set(kwargs.get('contains', []))
            kwargs['contains'].add(node.tag)

    tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    parent_xpath = '/'.join(tag_xpath.split('/')[:-1])

    #Which attributes are expected
    attribs = set()
    if parent_xpath in schema_dict['tag_info']:
        attribs = schema_dict['tag_info'][parent_xpath]['attribs']
        optional = schema_dict['tag_info'][parent_xpath]['optional_attribs']
    elif 'iteration_tag_info' in schema_dict:
        if parent_xpath in schema_dict['iteration_tag_info']:
            attribs = schema_dict['iteration_tag_info'][parent_xpath]['attribs']
            optional = schema_dict['iteration_tag_info'][parent_xpath]['optional_attribs']

    if only_required:
        attribs = attribs.difference(optional)

    if ignore is not None:
        attribs = attribs.difference(ignore)

    if not attribs:
        parser_info_out['parser_warnings'].append(f'Failed to evaluate attributes from parent tag of {name}: '
                                                  'No attributes to parse either the tag does not '
                                                  'exist or it has no attributes')
    else:
        attribs = sorted(list(attribs.original_case.values()))

    elems = eval_xpath(node, tag_xpath, parser_info_out=parser_info_out, list_return=True)

    out_dict = dict.fromkeys(attribs)
    for attrib in attribs:
        out_dict[attrib] = []

    for elem in elems:
        parent = elem.getparent()
        for attrib in attribs:

            stringattribute = get_xml_attribute(parent, attrib)

            if stringattribute == '':
                parser_info_out['parser_warnings'].append(
                    f'No values found for attribute {attrib} for parent tag of {name}')
                out_dict[attrib].append(None)
                continue

            possible_types = schema_dict['attrib_types'][attrib]

            warnings = []
            value, suc = convert_xml_attribute(stringattribute,
                                               possible_types,
                                               constants=constants,
                                               conversion_warnings=warnings)

            out_dict[attrib].append(value)

            if not suc:
                parser_info_out['parser_warnings'].append(f'Failed to evaluate attribute {attrib}: '
                                                          'Below are the warnings from convert_xml_attribute')
                for warning in warnings:
                    parser_info_out['parser_warnings'].append(warning)

    if all(len(x) == 1 for x in out_dict.values()):
        out_dict = {key: val[0] for key, val in out_dict.items()}

    return out_dict


@register_parsing_function('exists')
def tag_exists(node, schema_dict, name, parser_info_out=None, **kwargs):
    """
    Evaluates whether the tag exists in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: bool, True if any nodes with the path exist
    """
    return get_number_of_nodes(node, schema_dict, name, parser_info_out=parser_info_out, **kwargs) != 0


@register_parsing_function('numberNodes')
def get_number_of_nodes(node, schema_dict, name, parser_info_out=None, **kwargs):
    """
    Evaluates the number of occurences of the tag in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: bool, True if any nodes with the path exist
    """
    return len(eval_simple_xpath(node, schema_dict, name, parser_info_out=parser_info_out, list_return=True, **kwargs))


def eval_simple_xpath(node, schema_dict, name, parser_info_out=None, **kwargs):
    """
    Evaluates a simple xpath expression of the tag in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param list_return: bool, if True a list is always returned

    :returns: etree Elements obtained via the simple xpath expression
    """
    from masci_tools.util.xml.common_xml_util import eval_xpath

    list_return = kwargs.pop('list_return', False)

    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            kwargs['contains'] = set(kwargs.get('contains', []))
            kwargs['contains'].add(node.tag)

    tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    return eval_xpath(node, tag_xpath, parser_info_out=parser_info_out, list_return=list_return)



# -*- coding: utf-8 -*-
"""
Test of the functions in common_xml_util
"""
import pytest
import os
from masci_tools.util.constants import FLEUR_DEFINED_CONSTANTS
from pprint import pprint

TEST_FOLDER = os.path.dirname(os.path.abspath(__file__))
CLEAR_XML_TEST_FILE = os.path.abspath(os.path.join(TEST_FOLDER, 'files/fleur/test_clear.xml'))


def test_convert_to_float():
    """
    Test of the convert_to_float function
    """
    from masci_tools.util.xml.common_xml_util import convert_to_float

    ret_val, suc = convert_to_float('0.45')
    assert suc
    assert pytest.approx(ret_val) == 0.45

    ret_val, suc = convert_to_float('1.5324324e-9')
    assert suc
    assert pytest.approx(ret_val) == 1.5324324e-9

    warnings = []
    ret_val, suc = convert_to_float({}, conversion_warnings=warnings)
    assert not suc
    assert ret_val == {}
    assert warnings == ["Could not convert: '{}' to float, TypeError"]

    warnings = []
    ret_val, suc = convert_to_float('1,23', conversion_warnings=warnings)
    assert not suc
    assert ret_val == '1,23'
    assert warnings == ["Could not convert: '1,23' to float, ValueError"]

    warnings = []
    ret_val, suc = convert_to_float('.325352', conversion_warnings=warnings)
    assert suc
    assert pytest.approx(ret_val) == .325352
    assert warnings == []


def test_convert_to_int():
    """
    Test of the convert_to_int function
    """
    from masci_tools.util.xml.common_xml_util import convert_to_int

    ret_val, suc = convert_to_int('1241412')
    assert suc
    assert ret_val == 1241412

    ret_val, suc = convert_to_int('-9999999999999999999999')
    assert suc
    assert ret_val == -9999999999999999999999

    ret_val, suc = convert_to_int('12031')
    assert suc
    assert ret_val == 12031

    warnings = []
    ret_val, suc = convert_to_int((), conversion_warnings=warnings)
    assert not suc
    assert ret_val == ()
    assert warnings == ["Could not convert: '()' to int, TypeError"]

    warnings = []
    ret_val, suc = convert_to_int('1.231', conversion_warnings=warnings)
    assert not suc
    assert ret_val == '1.231'
    assert warnings == ["Could not convert: '1.231' to int, ValueError"]

    warnings = []
    ret_val, suc = convert_to_int('213', conversion_warnings=warnings)
    assert suc
    assert ret_val == 213
    assert warnings == []


def test_convert_from_fortran_bool():
    """
    Test of the convert_from_fortran_bool function
    """
    from masci_tools.util.xml.common_xml_util import convert_from_fortran_bool

    TRUE_ITEMS = ('T', 't', True)
    FALSE_ITEMS = ('F', 'f', False)

    for true_item in TRUE_ITEMS:
        ret_val, suc = convert_from_fortran_bool(true_item)
        assert suc
        assert ret_val

    for false_item in FALSE_ITEMS:
        ret_val, suc = convert_from_fortran_bool(false_item)
        assert suc
        assert not ret_val

    warnings = []
    ret_val, suc = convert_from_fortran_bool('TEST', conversion_warnings=warnings)
    assert not suc
    assert ret_val == 'TEST'
    assert warnings == ["Could not convert: 'TEST' to boolean, which is not 'True', 'False', 't', 'T', 'F' or 'f'"]

    warnings = []
    ret_val, suc = convert_from_fortran_bool({}, conversion_warnings=warnings)
    assert not suc
    assert ret_val == {}
    assert warnings == ["Could not convert: '{}' to boolean, only accepts str or boolean"]

    warnings = []
    ret_val, suc = convert_from_fortran_bool(True, conversion_warnings=warnings)
    assert suc
    assert ret_val
    assert warnings == []


def test_convert_to_fortran_bool():
    """
    Test of the function convert_to_fortran_bool
    """
    from masci_tools.util.xml.common_xml_util import convert_to_fortran_bool

    TRUE_ITEMS = (True, 'True', 't', 'T')
    FALSE_ITEMS = (False, 'False', 'f', 'F')

    for item in TRUE_ITEMS:
        res, suc = convert_to_fortran_bool(item)
        assert suc
        assert res == 'T'

    for item in FALSE_ITEMS:
        res, suc = convert_to_fortran_bool(item)
        assert suc
        assert res == 'F'

    warnings = []
    res, suc = convert_to_fortran_bool('True', conversion_warnings=warnings)
    assert suc
    assert res == 'T'
    assert warnings == []

    warnings = []
    res, suc = convert_to_fortran_bool('NOT_A_BOOL', conversion_warnings=warnings)
    assert not suc
    assert res is None
    assert warnings == [
        "A string: NOT_A_BOOL for a boolean was given, which is not 'True','False', 't', 'T', 'F' or 'f'"
    ]

    warnings = []
    res, suc = convert_to_fortran_bool((), conversion_warnings=warnings)
    assert not suc
    assert res is None
    assert warnings == ['convert_to_fortran_bool accepts only a string or bool as argument, given () ']


def test_eval_xpath():
    """
    Test of the eval_xpath function
    """
    from lxml import etree
    from masci_tools.util.xml.common_xml_util import eval_xpath

    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(CLEAR_XML_TEST_FILE, parser)
    root = xmltree.getroot()

    scfLoop = eval_xpath(root, '//scfLoop')
    assert isinstance(scfLoop, etree._Element)

    scfLoop = eval_xpath(root, '//scfLoop', list_return=True)
    assert len(scfLoop) == 1
    assert isinstance(scfLoop[0], etree._Element)

    include_tags = eval_xpath(root,
                              '//xi:include',
                              namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                              list_return=True)
    assert len(include_tags) == 2
    assert isinstance(include_tags[0], etree._Element)

    parser_info_out = {'parser_warnings': []}
    species_z = eval_xpath(root, "//species[@name='Cu-1']/@atomicNumber", parser_info_out=parser_info_out)
    assert species_z == '29'
    assert parser_info_out == {'parser_warnings': []}

    parser_info_out = {'parser_warnings': []}
    ldau_tags = eval_xpath(root, "//species[@name='Cu-1']/ldaU", parser_info_out=parser_info_out)
    assert ldau_tags == []
    assert parser_info_out == {'parser_warnings': []}

    parser_info_out = {'parser_warnings': []}
    ldau_tags = eval_xpath(root, "//species/[@name='Cu-1']/ldaU", parser_info_out=parser_info_out)
    assert ldau_tags == []
    assert parser_info_out == {
        'parser_warnings': [
            "There was a XpathEvalError on the xpath: //species/[@name='Cu-1']/ldaU \n Either it does not exist, or something is wrong with the expression."
        ]
    }


def test_clear_xml():
    """
    Test of the clear_xml function
    """
    from lxml import etree
    from masci_tools.util.xml.common_xml_util import eval_xpath, clear_xml
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(CLEAR_XML_TEST_FILE, parser)

    #Check that the file contains comments and includes
    root = xmltree.getroot()
    comments = eval_xpath(root, '//comment()', list_return=True)
    assert len(comments) == 3

    include_tags = eval_xpath(root,
                              '//xi:include',
                              namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                              list_return=True)
    assert len(include_tags) == 2

    symmetry_tags = eval_xpath(root, '//symOp', list_return=True)
    assert len(symmetry_tags) == 0

    cleared_tree = clear_xml(xmltree)
    cleared_root = cleared_tree.getroot()
    old_root = xmltree.getroot()

    #Make sure that the original tree was not modified
    comments = eval_xpath(old_root, '//comment()', list_return=True)
    assert len(comments) == 3

    #Check that the cleared tree is correct
    comments = eval_xpath(cleared_root, '//comment()', list_return=True)
    assert len(comments) == 0

    include_tags = eval_xpath(cleared_root,
                              '//xi:include',
                              namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                              list_return=True)
    assert len(include_tags) == 0

    symmetry_tags = eval_xpath(cleared_root, '//symOp', list_return=True)
    assert len(symmetry_tags) == 16


def test_get_xml_attribute():
    """
    Test of the clear_xml function
    """
    from lxml import etree
    from masci_tools.util.xml.common_xml_util import get_xml_attribute, eval_xpath
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(CLEAR_XML_TEST_FILE, parser)
    root = xmltree.getroot()

    scfLoop = eval_xpath(root, '//scfLoop')
    assert get_xml_attribute(scfLoop, 'alpha') == '.05000000'
    assert get_xml_attribute(scfLoop, 'itmax') == '1'

    parser_info_out = {'parser_warnings': []}
    assert get_xml_attribute(scfLoop, 'maxIterBroyd', parser_info_out=parser_info_out) == '99'
    assert parser_info_out == {'parser_warnings': []}

    parser_info_out = {'parser_warnings': []}
    assert get_xml_attribute(scfLoop, 'TEST', parser_info_out=parser_info_out) is None
    assert parser_info_out == {
        'parser_warnings': [
            'Tried to get attribute: "TEST" from element scfLoop.\n I recieved "None", maybe the attribute does not exist'
        ]
    }

    parser_info_out = {'parser_warnings': []}
    assert get_xml_attribute(xmltree, 'TEST', parser_info_out=parser_info_out) is None
    assert parser_info_out == {
        'parser_warnings': [
            'Can not get attributename: "TEST" from node of type <class \'lxml.etree._ElementTree\'>, because node is not an element of etree.'
        ]
    }


TEST_STRINGS = ['1.2314', 'all', ['all', '213', '-12'], ['PI', 'NOT_PI', '1.2'], ['F', 'T']]

TEST_TYPES = [['float'], ['int', 'string'], ['int', 'string'], ['float', 'float_expression'], ['int', 'switch']]

TEST_RESULTS = [(pytest.approx(1.2314), True), ('all', True), (['all', 213, -12], True),
                (['PI', 'NOT_PI', pytest.approx(1.2)], False), ([False, True], True)]

TEST_WARNINGS = [
    [], ["Could not convert: 'all' to int, ValueError"], ["Could not convert: 'all' to int, ValueError"],
    [
        "Could not convert: 'PI' to float, ValueError",
        "Could not evaluate expression 'PI' The following error was raised: Unknown string expression: PI",
        "Could not convert: 'NOT_PI' to float, ValueError",
        "Could not evaluate expression 'NOT_PI' The following error was raised: Unknown string expression: NOT"
    ], ["Could not convert: 'F' to int, ValueError", "Could not convert: 'T' to int, ValueError"]
]


@pytest.mark.parametrize('string_attr,types,results', zip(TEST_STRINGS, TEST_TYPES, TEST_RESULTS))
def test_convert_xml_attribute(string_attr, types, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_xml_attribute

    expected_val, expected_suc = results

    ret_val, suc = convert_xml_attribute(string_attr, types, FLEUR_DEFINED_CONSTANTS)
    assert ret_val == expected_val
    assert suc == expected_suc


@pytest.mark.parametrize('string_attr,types,results, warnings',
                         zip(TEST_STRINGS, TEST_TYPES, TEST_RESULTS, TEST_WARNINGS))
def test_convert_xml_attribute_warnings(string_attr, types, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_xml_attribute

    expected_val, expected_suc = results
    conversion_warnings = []
    ret_val, suc = convert_xml_attribute(string_attr,
                                         types,
                                         FLEUR_DEFINED_CONSTANTS,
                                         conversion_warnings=conversion_warnings)
    assert ret_val == expected_val
    assert suc == expected_suc
    assert conversion_warnings == warnings


TEST_ATTR_VALUES = [1.2134, 'all', ['all', 213, '-12'], ['3.14', 'NOT_PI', 1.2], [False, 'True']]

TEST_ATTR_TYPES = [['float'], ['int', 'string'], ['int', 'string'], ['float', 'float_expression'], ['int', 'switch']]

TEST_ATTR_RESULTS = [('1.2134000000', True), ('all', True), (['all', '213', '-12'], True),
                     (['3.14', 'NOT_PI', '1.2000000000'], True), (['F', 'T'], True)]

TEST_ATTR_WARNINGS = [
    [], [], [],
    [
        "Could not convert to float string '3.14' The following error was raised: Unknown format code 'f' for object of type 'str'",
        "Could not convert to float string '3.14' The following error was raised: Unknown format code 'f' for object of type 'str'",
        "Could not convert to float string 'NOT_PI' The following error was raised: Unknown format code 'f' for object of type 'str'",
        "Could not convert to float string 'NOT_PI' The following error was raised: Unknown format code 'f' for object of type 'str'"
    ], []
]


@pytest.mark.parametrize('attr_value,types,results', zip(TEST_ATTR_VALUES, TEST_ATTR_TYPES, TEST_ATTR_RESULTS))
def test_convert_attribute_to_xml(attr_value, types, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_attribute_to_xml

    expected_val, expected_suc = results

    ret_val, suc = convert_attribute_to_xml(attr_value, types)
    assert ret_val == expected_val
    assert suc == expected_suc


@pytest.mark.parametrize('attr_value,types,results,warnings',
                         zip(TEST_ATTR_VALUES, TEST_ATTR_TYPES, TEST_ATTR_RESULTS, TEST_ATTR_WARNINGS))
def test_convert_attribute_to_xml_warnings(attr_value, types, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_attribute_to_xml

    expected_val, expected_suc = results

    conversion_warnings = []
    ret_val, suc = convert_attribute_to_xml(attr_value, types, conversion_warnings=conversion_warnings)
    assert ret_val == expected_val
    assert suc == expected_suc
    assert conversion_warnings == warnings


TEST_TEXT_STRINGS = [
    '0.0 Pi/4.0 6.3121', '0.0 Pi/4.0 6.3121', '0.0 Pi/4.0 6.3121', '0.0 Pi/4.0 6.3121',
    ['0.0 Pi/4.0 6.3121', 'Bohr Pi/4.0 all', '0.0 Pi/*4.0 0.0'], ['F asd', 'T'],
    ['12 213 4215 412', '12 213 4215 412 123 14124']
]

TEST_DEFINITIONS = [[{
    'length': 3,
    'type': ['float', 'float_expression']
}], [{
    'length': 4,
    'type': ['float', 'float_expression']
}],
                    [{
                        'length': 4,
                        'type': ['float', 'float_expression']
                    }, {
                        'length': 'unbounded',
                        'type': ['float', 'float_expression']
                    }], [{
                        'length': 1,
                        'type': ['float', 'float_expression']
                    }], [{
                        'length': 3,
                        'type': ['float', 'float_expression']
                    }], [{
                        'length': 'unbounded',
                        'type': ['switch', 'string']
                    }], [{
                        'length': 4,
                        'type': ['switch', 'int']
                    }, {
                        'length': 'unbounded',
                        'type': ['int']
                    }]]

TEST_TEXT_RESULTS = [(pytest.approx([0.0, 0.7853981633974483, 6.3121]), True), ('0.0 Pi/4.0 6.3121', False),
                     (pytest.approx([0.0, 0.7853981633974483, 6.3121]), True), ('0.0 Pi/4.0 6.3121', False),
                     ([[0.0, 0.7853981633974483, 6.3121], [1.0, 0.7853981633974483, 'all'], [0.0, 'Pi/*4.0',
                                                                                             0.0]], False),
                     ([[False, 'asd'], [True]], True), ([[12, 213, 4215, 412], [12, 213, 4215, 412, 123, 14124]], True)]

TEST_TEXT_WARNINGS = [[], ["Failed to convert '0.0 Pi/4.0 6.3121', no matching definition found "], [],
                      [
                          "Could not convert: '0.0 Pi/4.0 6.3121' to float, ValueError",
                          "Could not evaluate expression '0.0 Pi/4.0 6.3121' The following error was "
                          'raised: Cannot parse number: Found two decimal points'
                      ],
                      [
                          "Could not convert: 'all' to float, ValueError",
                          "Could not evaluate expression 'all' The following error was raised: Unknown "
                          'string expression: all', "Could not convert: 'Pi/*4.0' to float, ValueError",
                          "Could not evaluate expression 'Pi/*4.0' The following error was raised: "
                          'Invalid Expression: Operator following operator'
                      ], [], []]


@pytest.mark.parametrize('string_text,definitions, results', zip(TEST_TEXT_STRINGS, TEST_DEFINITIONS,
                                                                 TEST_TEXT_RESULTS))
def test_convert_xml_text(string_text, definitions, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_xml_text

    expected_val, expected_suc = results

    ret_val, suc = convert_xml_text(string_text, definitions, FLEUR_DEFINED_CONSTANTS)
    assert ret_val == expected_val
    assert suc == expected_suc


@pytest.mark.parametrize('string_text,definitions,results,warnings',
                         zip(TEST_TEXT_STRINGS, TEST_DEFINITIONS, TEST_TEXT_RESULTS, TEST_TEXT_WARNINGS))
def test_convert_xml_text_warnings(string_text, definitions, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_xml_text

    expected_val, expected_suc = results

    conversion_warnings = []

    ret_val, suc = convert_xml_text(string_text,
                                    definitions,
                                    FLEUR_DEFINED_CONSTANTS,
                                    conversion_warnings=conversion_warnings)
    assert ret_val == expected_val
    assert suc == expected_suc
    assert conversion_warnings == warnings


TEST_TEXT_VALUES = [[0.0, 'Pi/4.0', 6.3121], [0.0, 'Pi/4.0', 6.3121], [0.0, 'Pi/4.0', 6.3121], [0.0, 'Pi/4.0', 6.3121],
                    [[0.0, 'Pi/4.0', 6.3121], ['Bohr', 'Pi/4.0', 'all']], [[False, 'asd'], 'T'],
                    [[12, '213', 4215, 412], [12, '213', 4215, '412', 123, 14124]]]

TEST_TEXT_XML_STRINGS = [(' 0.0000000000000 Pi/4.0  6.3121000000000', True), ('', False),
                         (' 0.0000000000000 Pi/4.0  6.3121000000000', True), ('', False),
                         ([' 0.0000000000000 Pi/4.0  6.3121000000000', 'Bohr Pi/4.0 all'], True), (['F asd',
                                                                                                    'T'], True),
                         (['12 213 4215 412', '12 213 4215 412 123 14124'], True)]

TEST_TEXT_XML_WARNINGS = [[], ["Failed to convert '[0.0, 'Pi/4.0', 6.3121]', no matching definition found "], [],
                          ["Failed to convert '[0.0, 'Pi/4.0', 6.3121]', no matching definition found "], [], [], []]


@pytest.mark.parametrize('text_value,definitions, results',
                         zip(TEST_TEXT_VALUES, TEST_DEFINITIONS, TEST_TEXT_XML_STRINGS))
def test_convert_text_to_xml(text_value, definitions, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_text_to_xml

    expected_val, expected_suc = results

    ret_val, suc = convert_text_to_xml(text_value, definitions)
    assert ret_val == expected_val
    assert suc == expected_suc


@pytest.mark.parametrize('text_value,definitions, results, warnings',
                         zip(TEST_TEXT_VALUES, TEST_DEFINITIONS, TEST_TEXT_XML_STRINGS, TEST_TEXT_XML_WARNINGS))
def test_convert_text_to_xml_warnings(text_value, definitions, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.common_xml_util import convert_text_to_xml

    expected_val, expected_suc = results

    conversion_warnings = []
    ret_val, suc = convert_text_to_xml(text_value, definitions, conversion_warnings=conversion_warnings)
    assert ret_val == expected_val
    assert suc == expected_suc
    assert conversion_warnings == warnings


def test_split_off_tag():
    """
    Test of the split_off_tag function
    """
    from masci_tools.util.xml.common_xml_util import split_off_tag

    assert split_off_tag('/fleurInput/calculationSetup/cutoffs') == ('/fleurInput/calculationSetup', 'cutoffs')
    assert split_off_tag('/fleurInput/calculationSetup/cutoffs/') == ('/fleurInput/calculationSetup', 'cutoffs')
    assert split_off_tag('./calculationSetup/cutoffs') == ('./calculationSetup', 'cutoffs')


def test_split_off_attrib():
    """
    Test of the split_off_tag function
    """
    from masci_tools.util.xml.common_xml_util import split_off_attrib

    assert split_off_attrib('/fleurInput/calculationSetup/cutoffs/@Kmax') == ('/fleurInput/calculationSetup/cutoffs',
                                                                              'Kmax')
    with pytest.raises(AssertionError):
        split_off_attrib('/fleurInput/calculationSetup/cutoffs')
    with pytest.raises(AssertionError):
        split_off_attrib("/fleurInput/atomSpecies/species[@name='TEST']")
    assert split_off_attrib('./calculationSetup/cutoffs/@Kmax') == ('./calculationSetup/cutoffs', 'Kmax')


def test_check_complex_xpath(load_inpxml):
    """
    Test of the check_complex_xpath function
    """
    from masci_tools.util.xml.common_xml_util import check_complex_xpath

    FILE_PATH = os.path.dirname(os.path.abspath(__file__))
    TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')

    xmltree, _ = load_inpxml(TEST_INPXML_PATH)

    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species', "/fleurInput/atomSpecies/species[@name='Fe-1']")

    with pytest.raises(ValueError):
        check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species',
                            "/fleurInput/atomSpecies/species[@name='Fe-1']/lo")

    with pytest.raises(ValueError):
        check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species',
                            "/fleurInput/atomSpecies/species[@name='Fe-1']/@name")

    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species', '/fleurInput/atomSpecies/species')
    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species',
                        "/fleurInput/atomSpecies/species[@name='does_not_exist']")
    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species/lo', "//species[@name='Pt-1']/lo")



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains a generic HDF5 reader
"""
import io
import h5py
from collections import namedtuple
import warnings

Transformation = namedtuple('Transformation', ['name', 'args', 'kwargs'])
AttribTransformation = namedtuple('AttribTransformation', ['name', 'attrib_name', 'args', 'kwargs'])


class HDF5Reader:
    """Class for reading in data from hdf5 files using a specified recipe

    :param file: filepath to hdf file or opened file handle (mode 'rb')
    :param move_to_memory: bool if True after reading and transforming the data
                           all leftover h5py.Datasets are moved into np.arrays

    The recipe is passed to the :py:meth:`HDF5Reader.read()` method and consists
    of a dict specifiying which attributes and datasets to read in and how to transform them

    Each attribute/dataset entry corresponds to one entry point in the given `.hdf` file
    Available transformations can either be found in :py:mod:`~masci_tools.io.parsers.hdf5.transforms`
    or can be defined by the user with the :py:func:`~masci_tools.io.parsers.hdf5.transforms.hdf5_transformation`
    decorator

    Basic Usage:

    .. code-block:: python

        from masci_tools.io.parsers.hdf5 import HDF5Reader
        import masci_tools.io.parsers.hdf5.recipes as recipes

        #This example shows the usage for producing data from a bandstructure calculation
        #in Fleur
        with HDF5Reader('/path/to/hdf/banddos.hdf') as h5reader:
            data, attributes = h5reader.read(recipe=recipes.FleurBands)
        print(data, attributes)

    """

    def __init__(self, file, move_to_memory=True):

        self._file = file

        if isinstance(self._file, io.IOBase):
            filename = self._file.name
        else:
            filename = self._file

        assert filename.endswith('.hdf'), f'Wrong File Type for {self.__class__}: Got {filename}'

        self._move_to_memory = move_to_memory
        self._h5_file = None

    def __enter__(self):
        self._h5_file = h5py.File(self._file, 'r')
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._h5_file.close()

    def _read_dataset(self, h5path, strict=True):
        """Return in the dataset specified by the given h5path

        :param h5path : str, HDF5 group path in file.
        :param strict : bool if no dataset is at the path and strcit is True raise Error
                        else return None

        :returns: h5py.Dataset or None

        :raises: ValueError if no dataset is at path and strict is True
        """
        if h5path in ('/', ''):
            if strict:
                return None
            else:
                pass

        dset = self._h5_file.get(h5path)
        if dset is not None:
            return dset
        elif strict:
            raise ValueError(f'HDF5 input file {self._file} has no Dataset at {h5path}.')
        return None

    def _transform_dataset(self, transforms, dataset, attributes=None):
        """
        Transforms the given dataset with the given list of tasks

        :param transforms: list of namedtuples defining the tasks to perform
        :param dataset: h5py.Dataset, on which to perform the operations
        :param attributes: dict of previously processed attributes.
                           Only available to the entries in the datasets
                           section of the recipe. This allows for operations with
                           the previously parsed attributes

        :returns: the dataset with all the transformations applied
        """
        transformed_dset = dataset
        for spec in transforms:

            args = spec.args
            if spec.name in self._attribute_transforms:
                if attributes is None:
                    raise ValueError('Attribute transform not allowed for attributes')
                attrib_value = attributes[spec.attrib_name]
                args = attrib_value, *args

            transformed_dset = self._transforms[spec.name](transformed_dset, *args, **spec.kwargs)

        return transformed_dset

    def read(self, recipe=None):
        """Extracts datasets from HDF5 file, transforms them and puts all into a namedtuple.

        :param recipe: dict with the format given in :py:mod:`~masci_tools.io.parsers.hdf5.recipes`

        :returns: two dicts with the datasets/attributes read in and transformed according to the recipe
        """
        from itertools import chain
        from masci_tools.io.hdf5_util import read_hdf_simple

        if recipe is None:
            warnings.warn('You are using the HDF5Reader without a recipe falling back to simple HDF reader')
            return read_hdf_simple(self._file)

        datasets = recipe.get('datasets', {})
        attributes = recipe.get('attributes', {})

        # remove entries whose key is an empty string
        h5paths = {item['h5path'] for item in chain(datasets.values(), attributes.values())}
        extracted_datasets = {h5path: self._read_dataset(h5path) for h5path in h5paths}

        output_attrs = {}
        for key, val in attributes.items():
            transforms = val.get('transforms', [])
            output_attrs[key] = self._transform_dataset(transforms, extracted_datasets[val['h5path']])
            if val.get('unpack_dict', False):
                if not isinstance(output_attrs[key], dict):
                    raise ValueError(f'{key} cannot be unpacked: Got {type(output_attrs[key])}')

                unpack_dict = output_attrs.pop(key)

                if unpack_dict.keys() & output_attrs.keys():
                    raise ValueError('Unpacking would result in lost information: \n'
                                     f"Intersection of keys: '{unpack_dict.keys().intersection(output_attrs.keys())}'")

                output_attrs = {**output_attrs, **unpack_dict}

        output_data = {}
        for key, val in datasets.items():
            transforms = val.get('transforms', [])
            output_data[key] = self._transform_dataset(transforms,
                                                       extracted_datasets[val['h5path']],
                                                       attributes=output_attrs)
            if val.get('unpack_dict', False):
                if not isinstance(output_data[key], dict):
                    raise ValueError(f'{key} cannot be unpacked: Got {type(output_data[key])}')

                unpack_dict = output_data.pop(key)
                if unpack_dict.keys() & output_data.keys():
                    raise ValueError('Unpacking would result in lost information: \n'
                                     f"Intersection of keys: '{unpack_dict.keys().intersection(output_data.keys())}'")

                output_data = {**output_data, **unpack_dict}

        if self._move_to_memory:
            self._transforms['move_to_memory'](output_data)
            self._transforms['move_to_memory'](output_attrs)

        return output_data, output_attrs



# -*- coding: utf-8 -*-
"""
Configurations for masci_tools tests
"""
import pytest


@pytest.fixture
def load_inpxml():
    """Returns the etree and schema_dict generator"""

    def _load_inpxml(path):
        from lxml import etree
        from masci_tools.io.parsers.fleur.fleur_schema import InputSchemaDict
        with open(path, 'r') as inpxmlfile:
            tree = etree.parse(inpxmlfile)
            version = str(tree.xpath('//@fleurInputVersion')[0])
            schema_dict = InputSchemaDict.fromVersion(version)
        return tree, schema_dict

    return _load_inpxml



# -*- coding: utf-8 -*-
"""
Tests for fleur outxml_parser specific conversion functions
"""
import pytest

TEST_DICTS = [{
    'end_date': {
        'date': '2020/12/10',
        'time': '16:51:35'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/10',
        'time': '16:58:49'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/10',
        'time': '22:31:67'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/11',
        'time': '16:51:35'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:35'
    }
}, {
    'end_date': {
        'date': '2020/12/31',
        'time': '09:12:45'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': None,
        'time': None
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/10',
        'time': '16:51:35'
    },
    'start_date': {
        'date': None,
        'time': None
    }
}, {
    'end_date': {
        'date': None,
        'time': None
    },
    'start_date': {
        'date': None,
        'time': None
    }
}]

TEST_WALLTIMES = [14, 448, 20446, 86400, 1786884, -60681, 60695, 0]

TEST_WARNINGS = [
    {
        'parser_warnings': []
    },
    {
        'parser_warnings': []
    },
    {
        'parser_warnings': []
    },
    {
        'parser_warnings': []
    },
    {
        'parser_warnings': []
    },
    {
        'parser_warnings': [
            'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!',
            'Enddate was unparsed, inp.xml prob not complete, do not believe the walltime!'
        ]
    },
    {
        'parser_warnings': [
            'Starttime was unparsed, inp.xml prob not complete, do not believe the walltime!',
            'Startdate was unparsed, inp.xml prob not complete, do not believe the walltime!'
        ]
    },
    {
        'parser_warnings': [
            'Starttime was unparsed, inp.xml prob not complete, do not believe the walltime!',
            'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!',
            'Startdate was unparsed, inp.xml prob not complete, do not believe the walltime!',
            'Enddate was unparsed, inp.xml prob not complete, do not believe the walltime!'
        ]
    },
]


@pytest.mark.parametrize('input_dict, walltime, warnings', zip(TEST_DICTS, TEST_WALLTIMES, TEST_WARNINGS))
def test_calculate_walltime(input_dict, walltime, warnings):
    """
   Test of the calculate_walltime function
   """
    from masci_tools.io.parsers.fleur.outxml_conversions import calculate_walltime

    parser_warnings = {'parser_warnings': []}
    out_dict = calculate_walltime(input_dict, parser_info_out=parser_warnings)

    assert out_dict['walltime_units'] == 'seconds'
    assert out_dict['walltime'] == walltime
    print(parser_warnings)
    assert parser_warnings == warnings



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains functions to load an fleur inp.xml file, parse it with a schema
and convert its content to a dict
"""
from lxml import etree
from pprint import pprint
from masci_tools.io.parsers.fleur.fleur_schema.schema_dict import InputSchemaDict
from masci_tools.util.xml.common_xml_util import clear_xml, validate_xml, convert_xml_attribute, convert_xml_text, eval_xpath
from masci_tools.util.schema_dict_util import read_constants


def inpxml_parser(inpxmlfile, version=None, parser_info_out=None):
    """
    Parses the given inp.xml file to a python dictionary utilizing the schema
    defined by the version number to validate and corretly convert to the dictionary

    :param inpxmlfile: either path to the inp.xml file, opened file handle or a xml etree to be parsed
    :param version: version string to enforce that a given schema is used
    :param parser_info_out: dict, with warnings, info, errors, ...

    :return: python dictionary with the parsed inp.xml

    :raises ValueError: If the validation against the schema failed, or an irrecoverable error
                        occured during parsing
    :raises FileNotFoundError: If no Schema file for the given version was found

    """

    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    parser_version = '0.2.0'
    parser_info_out['parser_info'] = f'Masci-Tools Fleur inp.xml Parser v{parser_version}'

    if isinstance(inpxmlfile, etree._ElementTree):
        xmltree = inpxmlfile
    else:
        parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
        try:
            xmltree = etree.parse(inpxmlfile, parser)
        except etree.XMLSyntaxError as msg:
            raise ValueError(f'Failed to parse input file: {msg}') from msg

    if version is None:
        version = eval_xpath(xmltree, '//@fleurInputVersion', parser_info_out=parser_info_out)
        version = str(version)
        if version is None:
            raise ValueError('Failed to extract inputVersion')

    parser_info_out['fleur_inp_version'] = version
    schema_dict = InputSchemaDict.fromVersion(version, parser_info_out=parser_info_out)

    ignore_validation = schema_dict['inp_version'] != version

    xmltree = clear_xml(xmltree)
    root = xmltree.getroot()

    constants = read_constants(root, schema_dict)

    try:
        validate_xml(xmltree, schema_dict.xmlschema, error_header='Input file does not validate against the schema')
    except etree.DocumentInvalid as err:
        errmsg = str(err)
        parser_info_out['parser_warnings'].append(errmsg)
        if not ignore_validation:
            raise ValueError(errmsg) from err

    if schema_dict.xmlschema.validate(xmltree) or ignore_validation:
        inp_dict = inpxml_todict(root, schema_dict, constants, parser_info_out=parser_info_out)
    else:
        parser_info_out['parser_warnings'].append('Input file does not validate against the schema: Reason is unknown')
        if not ignore_validation:
            raise ValueError('Input file does not validate against the schema: Reason is unknown')

    return inp_dict


def inpxml_todict(parent, schema_dict, constants, omitted_tags=False, base_xpath=None, parser_info_out=None):
    """
    Recursive operation which transforms an xml etree to
    python nested dictionaries and lists.
    Decision to add a list is if the tag name is in the given list tag_several

    :param parent: some xmltree, or xml element
    :param schema_dict: structure/layout of the xml file in python dictionary
    :param constants: dict with all the defined constants
    :param omitted_tags: switch. If True only a list of the contained tags is returned
                         Used to omitt useless tags like e.g ['atomSpecies']['species'][3]
                         becomes ['atomSpecies'][3]
    :param base_xpath: str, keeps track of the place in the inp.xml currently being processed
    :param parser_info_out: dict, with warnings, info, errors, ...

    :return: a python dictionary
    """

    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    #Check if this is the first call to this routine
    if base_xpath is None:
        base_xpath = f'/{parent.tag}'

    return_dict = {}
    if list(parent.items()):
        return_dict = dict(list(parent.items()))
        # Now we have to convert lazy fortan style into pretty things for the Database
        for key in return_dict:
            if key in schema_dict['attrib_types']:
                conversion_warnings = []
                return_dict[key], suc = convert_xml_attribute(return_dict[key],
                                                              schema_dict['attrib_types'][key],
                                                              constants,
                                                              conversion_warnings=conversion_warnings)
                if not suc:
                    parser_info_out['parser_warnings'].append(
                        f"Failed to convert attribute '{key}': "
                        'Below are the warnings raised from convert_xml_attribute')
                    for warning in conversion_warnings:
                        parser_info_out['parser_warnings'].append(warning)

    if parent.text:
        # has text, but we don't want all the '\n' s and empty stings in the database
        if parent.text.strip() != '':  # might not be the best solutions
            if parent.tag not in schema_dict['simple_elements']:
                raise ValueError(
                    f'Something is wrong in the schema_dict: {parent.tag} is not in simple_elements, but it has text')
            conversion_warnings = []
            converted_text, suc = convert_xml_text(parent.text,
                                                   schema_dict['simple_elements'][parent.tag],
                                                   constants,
                                                   conversion_warnings=conversion_warnings)
            if not suc:
                parser_info_out['parser_warnings'].append(f"Failed to convert text of '{parent.tag}': "
                                                          'Below are the warnings raised from convert_xml_text')
                for warning in conversion_warnings:
                    parser_info_out['parser_warnings'].append(warning)

            if not return_dict:
                return_dict = converted_text
            else:
                return_dict['text_value'] = converted_text
                if 'label' in return_dict:
                    return_dict['text_label'] = return_dict['label']
                    return_dict.pop('label')

    if base_xpath in schema_dict['tag_info']:
        tag_info = schema_dict['tag_info'][base_xpath]
    else:
        tag_info = {'several': []}

    for element in parent:

        new_base_xpath = f'{base_xpath}/{element.tag}'
        omitt_contained_tags = element.tag in schema_dict['omitt_contained_tags']
        new_return_dict = inpxml_todict(element,
                                        schema_dict,
                                        constants,
                                        base_xpath=new_base_xpath,
                                        omitted_tags=omitt_contained_tags,
                                        parser_info_out=parser_info_out)

        if element.tag in tag_info['several']:
            # make a list, otherwise the tag will be overwritten in the dict
            if element.tag not in return_dict:  # is this the first occurence?
                if omitted_tags:
                    if len(return_dict) == 0:
                        return_dict = []
                else:
                    return_dict[element.tag] = []
            if omitted_tags:
                return_dict.append(new_return_dict)
            elif 'text_value' in new_return_dict:
                for key, value in new_return_dict.items():
                    if key == 'text_value':
                        return_dict[element.tag].append(value)
                    elif key == 'text_label':
                        if 'labels' not in return_dict:
                            return_dict['labels'] = {}
                        return_dict['labels'][value] = new_return_dict['text_value']
                    else:
                        if key not in return_dict:
                            return_dict[key] = []
                        elif not isinstance(return_dict[key], list):  #Key seems to be defined already
                            raise ValueError(f'{key} cannot be extracted to the next level')
                        return_dict[key].append(value)
                for key in new_return_dict.keys():
                    if key in ['text_value', 'text_label']:
                        continue
                    if len(return_dict[key]) != len(return_dict[element.tag]):
                        raise ValueError(
                            f'Extracted optional argument {key} at the moment only label is supported correctly')
            else:
                return_dict[element.tag].append(new_return_dict)
        else:
            return_dict[element.tag] = new_return_dict

    return return_dict
